@@ -1,6 +1,6 @@
 use super::{
-    deep_map, deep_map_preorder, deep_map_unwrap_preorder, prelude::*, visit_circuit,
-    visit_circuit_fallable, HashBytes,
+    deep_map, deep_map_op_context_preorder_stoppable, deep_map_preorder, deep_map_unwrap,
+    deep_map_unwrap_preorder, prelude::*, visit_circuit, visit_circuit_fallable, HashBytes,
 };
 use crate::hashmaps::{AHashSet as HashSet, FxHashMap as HashMap};
 use crate::pycall;
@@ -47,10 +47,18 @@ pub fn replace_nodes_py(circuit: CircuitRc, map: HashMap<CircuitRc, CircuitRc>)
 /// Replaces child nodes recursively with a mapping
 /// In pre-order so that high level nodes are replaced first
 pub fn replace_nodes(circuit: CircuitRc, map: &HashMap<HashBytes, CircuitRc>) -> CircuitRc {
-    deep_map_unwrap_preorder(&circuit, |x: &Circuit| -> CircuitRc {
-        let rc = x.clone().rc();
-        map.get(&rc.info().hash).cloned().unwrap_or(rc)
-    })
+    deep_map_op_context_preorder_stoppable(
+        &circuit,
+        &|x: &Circuit, _| {
+            let rc = x.clone().rc();
+            let result = map.get(&rc.info().hash).cloned();
+            let stop = result.is_some();
+            (result, stop)
+        },
+        &mut (),
+        &mut Default::default(),
+    )
+    .unwrap_or(circuit)
 }
 
 #[pyfunction]
