@@ -1,6 +1,6 @@
 use std::{collections::BTreeSet, iter::zip};
 
-use crate::hashmaps::FxHashMap as HashMap;
+use crate::hashmaps::{AHashSet as HashSet, FxHashMap as HashMap};
 
 use crate::{
     opt_einsum::EinsumSpec,
@@ -10,10 +10,12 @@ use crate::{
 use num_bigint::BigUint;
 
 use super::{
-    apply_fn_cache, deep_map_op_pre_new_children, deep_map_unwrap, visit_circuit, ArrayConstant,
-    Circuit, CircuitNode, CircuitRc, HashBytes, Index,
+    apply_fn_cache, deep_map_op_pre_new_children, deep_map_unwrap, parse_compiler_repr_bijection,
+    visit_circuit, visit_circuit_fallable, visit_circuit_with_parents, ArrayConstant, Circuit,
+    CircuitConstructionError, CircuitNode, CircuitRc, HashBytes, Index,
 };
 use crate::tensor_util::TorchDeviceDtype;
+use std::hash::Hash;
 
 #[pyfunction]
 pub fn cast_circuit(circ: CircuitRc, device_dtype: &TorchDeviceDtype) -> CircuitRc {
@@ -79,6 +81,14 @@ pub fn hash_to_node(circuit: &Circuit) -> HashMap<HashBytes, CircuitRc> {
     result
 }
 
+pub fn subcircuits(circuit: &Circuit) -> HashSet<CircuitRc> {
+    let mut result: HashSet<CircuitRc> = HashSet::new();
+    visit_circuit(circuit, &mut |x: &Circuit| {
+        result.insert(x.clone().rc());
+    });
+    result
+}
+
 #[pyfunction]
 pub fn total_flops(circuit: CircuitRc) -> BigUint {
     let mut result: BigUint = BigUint::from(0usize);
@@ -152,6 +162,7 @@ pub fn get_all_einsum_specs(circuit: CircuitRc) -> Vec<EinsumSpec> {
 }
 
 #[pyfunction]
+/// children first
 pub fn toposort_circuit(circuit: CircuitRc) -> Vec<CircuitRc> {
     let mut num_refs: HashMap<CircuitRc, usize> = HashMap::new();
     visit_circuit(&circuit, |c| {
@@ -193,6 +204,85 @@ where
     panic!();
 }
 
+pub fn deep_map_pass_down<F, F2, T>(circuit: CircuitRc, pass_down: F, make_circuit: F2) -> CircuitRc
+where
+    F: Fn(CircuitRc, &Vec<T>) -> Vec<T>,
+    F2: Fn(CircuitRc, &Vec<T>) -> CircuitRc,
+{
+    let mut toposorted = toposort_circuit(circuit.clone());
+    toposorted.reverse();
+    let mut passed_down: HashMap<CircuitRc, Vec<T>> = HashMap::new();
+    for c in toposorted {
+        let pass = pass_down(c.clone(), passed_down.get(&c).unwrap_or(&vec![]));
+        for (child, pass) in zip(c.children(), pass) {
+            passed_down
+                .entry(child.clone())
+                .or_insert(vec![])
+                .push(pass);
+        }
+    }
+    deep_map_unwrap(&circuit, |c| {
+        make_circuit(
+            c.clone().rc(),
+            passed_down.get(&c.clone().rc()).unwrap_or(&vec![]),
+        )
+    })
+}
+
+pub fn deep_map_pass_down_branching<F, F2, T>(
+    circuit: CircuitRc,
+    pass_down_f: F,
+    make_circuit_f: F2,
+    initial_pass_down: T,
+) -> CircuitRc
+where
+    T: Hash + Eq + Clone,
+    F: Fn(CircuitRc, &T) -> Vec<T>,
+    F2: Fn(CircuitRc, &T, &Vec<CircuitRc>) -> CircuitRc,
+{
+    let mut pass_down_cache: HashMap<(CircuitRc, T), Vec<T>> = HashMap::new();
+    let mut construct_cache: HashMap<(CircuitRc, T), CircuitRc> = HashMap::new();
+
+    fn recurse<F, F2, T>(
+        circuit: CircuitRc,
+        passed: T,
+        pdf: &F,
+        mcf: &F2,
+        pdc: &mut HashMap<(CircuitRc, T), Vec<T>>,
+        cc: &mut HashMap<(CircuitRc, T), CircuitRc>,
+    ) -> CircuitRc
+    where
+        T: Hash + Eq + Clone,
+        F: Fn(CircuitRc, &T) -> Vec<T>,
+        F2: Fn(CircuitRc, &T, &Vec<CircuitRc>) -> CircuitRc,
+    {
+        let pass_key = (circuit.clone(), passed.clone());
+        if let Some(result) = cc.get(&pass_key) {
+            return result.clone();
+        }
+        let passing = pdc.get(&pass_key).cloned().unwrap_or_else(|| {
+            let result = pdf(circuit.clone(), &passed);
+            pdc.insert(pass_key.clone(), result.clone());
+            result
+        });
+        let new_children: Vec<CircuitRc> = zip(circuit.children(), passing)
+            .map(|(child, pass)| recurse(child, pass, pdf, mcf, pdc, cc))
+            .collect();
+
+        let result = mcf(circuit, &passed, &new_children);
+        cc.insert(pass_key, result.clone());
+        result
+    }
+    recurse(
+        circuit,
+        initial_pass_down,
+        &pass_down_f,
+        &make_circuit_f,
+        &mut pass_down_cache,
+        &mut construct_cache,
+    )
+}
+
 #[pyfunction]
 pub fn replace_all_randn_seeded(circuit: CircuitRc) -> CircuitRc {
     deep_map_unwrap(&circuit, |node| match node {
@@ -227,23 +317,25 @@ pub fn replace_all_randn_seeded(circuit: CircuitRc) -> CircuitRc {
     })
 }
 
-// pub fn deep_map_pass_down<F, T>(circuit: &Circuit, f: &mut F) -> CircuitRc
-// where
-//     T: Clone,
-//     F: FnMut(&Circuit, Vec<T>) -> (CircuitRc, Vec<T>),
-// {
-//     let mut topo = toposort_circuit(circuit);
-//     topo.reverse();
-//     let mut pass_downs: HashMap<CircuitRc, Vec<T>> = HashMap::new();
-//     for c in topo {
-//         let passed_down = pass_downs.remove(&c).unwrap();
-//         let (new, pass_here) = f(&c, passed_down);
-//         for (child,passed) in zip(c.children(),pass_here){
-//             pass_downs.entry(child).or_insert(vec![]).push(passed);
-//         }
-//         if &**c == circuit {
-//             return (new, pass_here);
-//         }
-//     }
-//     panic!();
-// }
+pub fn has_leaf_symbols(circuit: CircuitRc, allowed: &HashSet<HashBytes>) -> bool {
+    let mut result = false;
+    visit_circuit(&circuit, |circuit| {
+        if circuit.as_symbol().is_some() && !allowed.contains(&circuit.info().hash) {
+            println!(
+                "found symbol {:?} allowed {:?}",
+                circuit.info().hash,
+                allowed
+            );
+            result = true;
+        }
+        if let Circuit::ModuleNode(mn) = circuit {
+            let inner_allowed: HashSet<HashBytes> = allowed
+                .iter()
+                .cloned()
+                .chain(mn.spec.input_specs.iter().map(|s| s.symbol.info().hash))
+                .collect();
+            result = result || has_leaf_symbols(mn.spec.spec_circuit.clone(), &inner_allowed);
+        }
+    });
+    result
+}
