@@ -1,4 +1,5 @@
 #![allow(clippy::borrow_deref_ref)]
+use crate::circuit::circuit_utils::toposort_circuit;
 use crate::hashmaps::{AHashSet as HashSet, FxHashMap as HashMap};
 use crate::pyo3_prelude::*;
 use crate::{
@@ -23,6 +24,7 @@ use num_bigint::BigUint;
 pub use parsing::parse_compiler_repr_bijection;
 use py_circuit_items::circuit_rust_to_py;
 use pyo3::{exceptions, pyclass::CompareOp};
+use std::collections::BTreeMap;
 use std::hash::Hash;
 use std::{
     iter::zip,
@@ -75,7 +77,7 @@ use self::{
     print::{print_circuit_stats, repr_circuit_deep_compiler},
 };
 
-pub type NamedAxes = Option<Vec<Option<String>>>;
+pub type NamedAxes = BTreeMap<u8, String>;
 pub trait CircuitNode: CircuitNodePrivate {
     // ==== implementable section ===
     //
@@ -150,26 +152,22 @@ pub trait CircuitNode: CircuitNodePrivate {
     }
 
     fn compute_named_axes(&self) -> NamedAxes {
-        if !self.children().any(|x| x.info().named_axes.is_some()) {
-            return None;
+        if !self.info().named_axes.is_empty() {
+            return self.info().named_axes.clone();
+        }
+        if !self.children().any(|x| !x.info().named_axes.is_empty()) {
+            return BTreeMap::new();
         }
         let child_axis_map = self.child_axis_map();
-        let mut result: Vec<Option<String>> = vec![None; self.info().rank()];
-        let mut did_anything = false;
+        let mut result: NamedAxes = BTreeMap::new();
         for (mp, child) in zip(child_axis_map, self.children()) {
-            if let Some(child_names) = &child.info().named_axes {
-                for (out_axis, name) in zip(mp, child_names) {
-                    if let Some(out_axis)=out_axis && let Some(name)=name{
-                    result[out_axis]=Some(name.clone());
-                    did_anything=true;
-                }
+            for (ax, name) in &child.info().named_axes {
+                if let Some(top_ax) = mp[(*ax) as usize] {
+                    result.insert(top_ax as u8, name.clone());
                 }
             }
         }
-        if !did_anything {
-            return None;
-        }
-        Some(result)
+        result
     }
 
     fn map_children<'a, F, E>(&'a self, mut f: F) -> Result<Self, CircuitConstructionError>
@@ -247,14 +245,16 @@ pub trait CircuitNode: CircuitNodePrivate {
         self.info_mut().shape = shape.clone(); // set shape so methods to compute other info can use it
 
         let mut hasher = self.compute_hash();
-        let named_axes = self.compute_named_axes();
+        self.info_mut().named_axes = self.compute_named_axes();
         hasher.update(self.name().unwrap_or("").as_bytes());
         hasher.update(&self.node_type_uuid());
         hasher.update(uuid!("025e9af4-1366-4211-aa5f-7c28fc6cdf9f").as_bytes());
-        if let Some(named_axes_real) = &named_axes {
-            for axis in named_axes_real {
-                hasher.update(axis.as_ref().unwrap_or(&"".to_owned()).as_bytes());
+        for (axis, name) in &self.info().named_axes {
+            if *axis as usize >= shape.len() {
+                return Err(CircuitConstructionError::NamedAxisAboveRank {});
             }
+            hasher.update(&[*axis]);
+            hasher.update(name.as_bytes());
         }
         self.info_mut().is_constant = self.compute_is_constant();
         self.info_mut().is_explicitly_computable = self.compute_is_explicitly_computable();
@@ -262,7 +262,6 @@ pub trait CircuitNode: CircuitNodePrivate {
         self.info_mut().hash = hasher.finalize().into();
         self.info_mut().max_non_input_size = self.max_non_input_size();
         self.info_mut().device_dtype = self.compute_device_dtype()?;
-        self.info_mut().named_axes = named_axes;
         Ok(self)
     }
 
@@ -574,9 +573,39 @@ macro_rules! define_circuit_union {
                 }
             }
         }
+        impl ::std::convert::From<$crate::circuit::Circuit> for Option<$name> {
+            fn from(item: $crate::circuit::Circuit) -> Option<$name> {
+                match item {
+                    $(
+                        $crate::circuit::Circuit::$x(node) => Some(node.into()),
+                    )*
+                    _=>None
+                }
+            }
+        }
+        impl $name{
+            pub fn matches(circuit:&$crate::circuit::Circuit)->bool{
+                let op:Option<$name>=circuit.clone().into();
+                op.is_some()
+            }
+        }
     }
 }
 
+// These nodes are uneffected by rewrites, and satisfy
+// AlgebraicRewrite(Replace(X, IrreducibleNode->Y)) == Replace(AlgebraicRewrite(IrreducibleNode), IrreducibleNode->Y)
+// except for hashmap iteration order or other unfortunate nondeterminism
+define_circuit_union!(IrreducibleNode {
+    ArrayConstant,
+    Symbol,
+});
+
+define_circuit_union!(Leaf {
+    ArrayConstant,
+    Symbol,
+    ScalarConstant,
+});
+
 // work around for fact that we can't implement foreign trait on constrained type
 #[macro_export]
 macro_rules! circuit_node_extra_impl {
@@ -836,6 +865,9 @@ pub enum CircuitConstructionError {
     #[error("Scatter not supported yet, {index:?}")]
     ScatterUnimplemented { index: TensorIndex },
 
+    #[error("num_batches {num_batches} doesn't divide length {l}")]
+    BatchNumberDoesntDivide { l: usize, num_batches: usize },
+
     #[error("Children multiple dtypes {a:?} {b:?}")]
     ChildrenMultipleDtypes {
         a: Option<String>,
@@ -887,8 +919,11 @@ pub enum CircuitConstructionError {
         new_shape: Shape,
     },
 
-    #[error("Inputs that should have same batching have different batchings, {batch_shapes:?}")]
-    InconsistentBatches { batch_shapes: Vec<Shape> },
+    #[error("Inputs that should have same batching have different batchings, {batch_shapes:?} {circuit:?}")]
+    InconsistentBatches {
+        batch_shapes: Vec<Shape>,
+        circuit: CircuitRc,
+    },
 
     #[error("ModuleNode got unknown keyword argument, {argument}")]
     ModuleNodeUnknownArgument { argument: String },
@@ -923,6 +958,27 @@ pub enum CircuitConstructionError {
     #[error("Parsing found multiple circuits in sequence, instead of one circuit (aka last node wasn't child of anything)")]
     ParsingGotMultipleCircuits {},
 
+    #[error("Named axis higher than rank")]
+    NamedAxisAboveRank {},
+
+    #[error("trying to expand node, unknown variant {variant}")]
+    ExpandNodeUnknownVariant { variant: String },
+
+    #[error("Batching axis originates too high")]
+    BatchAxisOriginatesTooHigh {},
+
+    #[error("ExpandingRemovableAxisUnfortunateError")]
+    ExpandingRemovableAxisUnfortunateError {},
+
+    #[error("Bug Error (yikes!) (aka unreachable code / assert failure)")]
+    BugError {},
+
+    #[error("Module by name not found: {name}")]
+    ModuleNotFound { name: String },
+
+    #[error("Circuit by name not found: {name}")]
+    CircuitRefNotFound { name: String },
+
     #[error("python error {py_err:?}")]
     PythonError { py_err: Rc<PyErr> },
 }
@@ -1081,6 +1137,42 @@ where
     recurse(circuit, &mut seen, &mut f);
 }
 
+pub fn visit_circuit_with_parents<F>(circuit: &Circuit, mut f: F)
+where
+    F: FnMut(&Circuit, &Vec<CircuitRc>),
+{
+    let mut toposorted = toposort_circuit(circuit.clone().rc());
+    toposorted.reverse(); // its children first by default
+
+    let mut parents: HashMap<CircuitRc, Vec<CircuitRc>> = HashMap::new();
+    for (i, sub) in toposorted.into_iter().enumerate() {
+        f(&sub, &parents.get(&sub).unwrap_or(&vec![]));
+        for child in sub.children() {
+            parents.entry(child).or_insert(vec![]).push(sub.clone());
+        }
+    }
+}
+
+pub fn visit_circuit_with_parents_fallible<F, E>(circuit: &Circuit, mut f: F) -> Result<(), E>
+where
+    F: FnMut(&Circuit, &Vec<CircuitRc>) -> Result<(), E>,
+{
+    let mut toposorted = toposort_circuit(circuit.clone().rc());
+    toposorted.reverse(); // its children first by default
+
+    let mut parents: HashMap<CircuitRc, Vec<CircuitRc>> = HashMap::new();
+    for (i, sub) in toposorted.into_iter().enumerate() {
+        f(&sub, &parents.get(&sub).unwrap_or(&vec![]))?;
+        for child in sub.children() {
+            parents.entry(child).or_insert(vec![]).push(sub.clone());
+        }
+    }
+    Ok(())
+}
+
+/// does not visit children of circuits where f fails. It does visit all children even if one fails
+/// even though this is more work than stopping on the first child that fails
+/// because it's semantically cleaner to not have to think about which children are first
 pub fn visit_circuit_fallable<F, E>(circuit: &Circuit, mut f: F) -> Result<(), E>
 where
     F: FnMut(&Circuit) -> Result<(), E>,
@@ -1094,9 +1186,12 @@ where
         if !seen.contains(&circ.info().hash) {
             seen.insert(circ.info().hash);
             f(circ)?;
-            for child in circ.children() {
-                recurse(&child, seen, f)?
-            }
+
+            circ.children()
+                .map(|child| recurse(&child, seen, f))
+                .collect::<Vec<_>>() // intermediate collect causes all recurses to happen even if one errors
+                .into_iter()
+                .collect::<Result<Vec<_>, _>>()?;
         }
         Ok(())
     }
@@ -1188,7 +1283,7 @@ where
     #[apply(cached_lambda)]
     #[key(circ.info().hash, HashBytes)]
     fn recurse(circ: &Circuit) -> Result<CircuitRc, CircuitConstructionError> {
-        let old_children: Vec<CircuitRc> = circ.children().collect();
+        let old_children: Vec<CircuitRc> = circ.children().collect(); // need to define this for borrow reasons
         let new_children: Result<Vec<CircuitRc>, CircuitConstructionError> =
             old_children.iter().map(|x| &***x).map(recurse).collect();
         new_children.and_then(|a| f(circ, &a))
@@ -1235,6 +1330,40 @@ where
     result
 }
 
+pub fn deep_map_op_context_preorder_stoppable<F, C>(
+    circuit: &Circuit,
+    f: &F,
+    context: &mut C,
+    self_cache: &mut HashMap<HashBytes, Option<CircuitRc>>,
+) -> Option<CircuitRc>
+where
+    F: Fn(&Circuit, &mut C) -> (Option<CircuitRc>, bool),
+{
+    if let Some(z) = self_cache.get(&circuit.info().hash) {
+        return z.clone();
+    }
+    let (circuit_applied, stop) = f(circuit, context);
+    if stop {
+        return circuit_applied;
+    }
+    let result = if let Some(applied) = circuit_applied {
+        Some(
+            applied
+                .map_children_op(|x| {
+                    deep_map_op_context_preorder_stoppable(x, f, context, self_cache)
+                })
+                .map(|x| x.rc())
+                .unwrap_or(applied.clone()),
+        )
+    } else {
+        circuit
+            .map_children_op(|x| deep_map_op_context_preorder_stoppable(x, f, context, self_cache))
+            .map(|x| x.rc())
+    };
+    self_cache.insert(circuit.info().hash, result.clone());
+    result
+}
+
 pub fn evaluate_fn_uncached(
     circ: &Circuit,
     device_dtype: &TorchDeviceDtype,
@@ -1438,6 +1567,11 @@ impl PyCircuitBase {
         self.0.info().rank()
     }
 
+    #[pyo3(name = "child_axis_map")]
+    fn child_axis_map_py(&self) -> Vec<Vec<Option<usize>>> {
+        self.child_axis_map()
+    }
+
     fn to_py(&self) -> PyObject {
         circuit_rust_to_py(CircuitRc(self.0.clone()))
     }
