@@ -1,6 +1,7 @@
 use crate::{
     hashmaps::{AHashSet as HashSet, FxHashMap as HashMap},
     py_types::{scalar_to_tensor, tensor_scale, un_flat_concat, PY_UTILS},
+    timed_value,
 };
 use std::fmt::{self, Display};
 
@@ -16,10 +17,12 @@ use super::{
     circuit_optimizer::{OptimizationContext, OptimizationSettings},
     circuit_utils::{hash_to_node, toposort_circuit},
     get_compatible_dtype,
+    module_nodes::inline_all_modules,
     prelude::*,
     print::oom_fmt,
     scheduling_z3::schedule_dag_strategy_ints,
-    visit_circuit, ArrayConstant, HashBytes, PyHashBytes, ScalarConstant,
+    visit_circuit, Add, ArrayConstant, HashBytes, IrreducibleNode, Leaf, ModuleNodeSpec,
+    PyHashBytes, ScalarConstant, Symbol,
 };
 use crate::pyo3_prelude::*;
 use crate::util::filter_to_idx;
@@ -31,37 +34,45 @@ use thiserror::Error;
 
 #[derive(Clone, Debug)]
 pub enum Instruction {
-    Drop(HashBytes),
-    Compute(CircuitRc),
+    Drop(usize),
+    Compute(usize, CircuitRc),
 }
 
 impl IntoPy<PyObject> for Instruction {
     fn into_py(self, py: Python<'_>) -> PyObject {
         match self {
-            Instruction::Drop(hb) => PyHashBytes(hb).into_py(py),
-            Instruction::Compute(c) => c.into_py(py),
+            Instruction::Drop(hb) => hb.into_py(py),
+            Instruction::Compute(key, circ) => (key, circ).into_py(py),
         }
     }
 }
 
+/// Schedule is an optimized sequence of operations to perform to compute a circuit
+/// it keeps arrayconstants seperately, and replacing arrayconstants in a schedule should produce the same
+/// output + ~performance as optimizing the circuit with different arrayconstants from the beginning
+/// Schedule stores intermediate nodes as circuits with Symbol children, where the name of the symbol
+/// contains the "id" of that node
 #[pyclass(unsendable)]
 #[derive(Clone, Debug, PyClassDeriv)]
 pub struct Schedule {
     #[pyo3(get)]
     pub instructions: Vec<Instruction>,
-    pub constants: HashMap<HashBytes, Tensor>,
+    #[pyo3(get)]
+    pub constants: HashMap<usize, IrreducibleNode>,
     // keep scalar constants seperate so adjust numerical scale can work without losing precision
     // before when these were in tensors, they had to be right dtype and therefore overflowed when
     // adjustment needed
-    pub scalar_constants: HashMap<HashBytes, ScalarConstant>,
+    #[pyo3(get)]
+    pub scalar_constants: HashMap<usize, ScalarConstant>,
     #[pyo3(get)]
     pub device_dtype: TorchDeviceDtype,
-    pub output_circuit: Option<CircuitRc>,
+    pub output_circuit: Option<(usize, CircuitRc)>,
     pub split_shapes: Option<Vec<Shape>>,
 }
+
 #[pymethods]
 impl Schedule {
-    pub fn replace_tensors(&self, map: HashMap<HashBytes, Tensor>) -> PyResult<Self> {
+    pub fn replace_tensors(&self, map: HashMap<usize, Tensor>) -> PyResult<Self> {
         let mut result = self.clone();
         for (k, v) in map {
             if !result.constants.contains_key(&k) {
@@ -69,7 +80,9 @@ impl Schedule {
                     "key circuit wasn't present in original",
                 ));
             }
-            result.constants.insert(k, v);
+            result
+                .constants
+                .insert(k, ArrayConstant::new(v, None).into());
         }
         Ok(result)
     }
@@ -80,28 +93,19 @@ impl Schedule {
         result.constants = result
             .constants
             .into_iter()
-            .map(|(hash, tensor)| (hash, pycall!(f, (PyHashBytes(hash), tensor))))
+            .map(|(key, tensor)| (key, pycall!(f, (key, tensor))))
             .collect();
         Ok(result)
     }
 
-    #[getter]
-    #[pyo3(name = "constants")]
-    pub fn constants_py(&self) -> HashMap<PyHashBytes, Tensor> {
-        self.constants
-            .iter()
-            .map(|(k, v)| (PyHashBytes(*k), v.clone()))
-            .collect()
-    }
-
     pub fn evaluate(&self, settings: OptimizationSettings) -> Tensor {
         let eval = || {
             let result = if settings.adjust_numerical_scale {
                 evaluate_schedule_adjust_numerical_scale(self, settings)
-                    [&self.output_circuit.as_ref().unwrap().info().hash]
+                    [&self.output_circuit.as_ref().unwrap().0]
                     .clone()
             } else {
-                evaluate_schedule(self)[&self.output_circuit.as_ref().unwrap().info().hash].clone()
+                evaluate_schedule(self)[&self.output_circuit.as_ref().unwrap().0].clone()
             };
             if self.device_dtype.device != "cpu".to_owned() {
                 Python::with_gil(|py| {
@@ -132,17 +136,17 @@ impl Schedule {
     pub fn get_stats(&self) -> ScheduleStats {
         let mut mem: BigUint = BigUint::from(0usize);
         let mut max_mem: BigUint = BigUint::from(0usize);
-        let mut biggest: HashMap<HashBytes, CircuitRc> = HashMap::new();
-        let mut current: HashMap<HashBytes, CircuitRc> = HashMap::new();
+        let mut biggest: HashMap<usize, CircuitRc> = HashMap::new();
+        let mut current: HashMap<usize, CircuitRc> = HashMap::new();
         for instruction in self.instructions.clone() {
             match instruction {
                 Instruction::Drop(drop) => {
                     let dropped = current.remove(&drop).unwrap();
                     mem -= dropped.info().numel();
                 }
-                Instruction::Compute(c) => {
-                    current.insert(c.info().hash, c.clone());
-                    mem += c.info().numel();
+                Instruction::Compute(key, circuit) => {
+                    current.insert(key, circuit.clone());
+                    mem += circuit.info().numel();
                     if mem > max_mem {
                         max_mem = mem.clone();
                         biggest = current.clone();
@@ -155,11 +159,68 @@ impl Schedule {
             constant_mem: self
                 .constants
                 .iter()
-                .map(|(_h, t)| t.shape().iter().product::<usize>())
+                .map(|(_h, t)| t.info().shape.iter().product::<usize>())
                 .sum(),
             max_circuit_set: biggest.values().cloned().collect(),
         }
     }
+
+    pub fn next_key(&self) -> usize {
+        *self
+            .constants
+            .keys()
+            .chain(self.scalar_constants.keys())
+            .chain(self.instructions.iter().filter_map(|ins| match ins {
+                Instruction::Compute(k, c) => Some(k),
+                _ => None,
+            }))
+            .max()
+            .unwrap_or(&0)
+            + 1
+    }
+}
+
+impl Display for Schedule {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        write!(
+            f,
+            "Schedule: instructions\n{}\nArrays: {} Symbols: {} Scalars: {}",
+            self.instructions
+                .iter()
+                .filter_map(|i| {
+                    if let Instruction::Compute(k, c) = i {
+                        return Some(
+                            k.to_string()
+                                + " "
+                                + &c.variant_string()
+                                + " "
+                                + &c.children()
+                                    .map(|x| get_child_key(x).to_string())
+                                    .collect::<Vec<String>>()
+                                    .join(" "),
+                        );
+                    }
+                    None
+                })
+                .collect::<Vec<String>>()
+                .join("\n"),
+            self.constants
+                .iter()
+                .filter_map(|(k, c)| c.as_array_constant().map(|_| k.to_string()))
+                .collect::<Vec<String>>()
+                .join(" "),
+            self.constants
+                .iter()
+                .filter_map(|(k, c)| c.as_symbol().map(|_| k.to_string()))
+                .collect::<Vec<String>>()
+                .join(" "),
+            self.scalar_constants
+                .iter()
+                .map(|(k, c)| k.to_string())
+                .collect::<Vec<String>>()
+                .join(" "),
+        )
+    }
 }
 
 #[pyclass(unsendable)]
@@ -184,6 +245,9 @@ pub enum SchedulingError {
     },
     #[error("Single element doesn't fit {numel}")]
     SingleElementTooBig { numel: BigUint },
+
+    #[error("Circuit has leaf symbols")]
+    LeafSymbols {},
 }
 
 create_exception!(
@@ -230,8 +294,34 @@ impl Display for ScheduleStats {
     }
 }
 
-pub fn evaluate_schedule(schedule: &Schedule) -> HashMap<HashBytes, Tensor> {
-    let mut live: HashMap<HashBytes, Tensor> = schedule.constants.clone();
+fn get_child_key(circuit: CircuitRc) -> usize {
+    circuit
+        .as_symbol()
+        .unwrap()
+        .name()
+        .unwrap()
+        .parse::<usize>()
+        .unwrap()
+}
+
+fn child_from_key(key: usize, shape: Shape) -> Symbol {
+    Symbol::new_with_random_uuid(shape, Some(key.to_string()))
+}
+
+fn identity(key: usize, shape: Shape) -> CircuitRc {
+    Add::nrc(
+        vec![Symbol::new_with_random_uuid(shape, Some(key.to_string())).rc()],
+        None,
+    )
+}
+
+pub fn evaluate_schedule(schedule: &Schedule) -> HashMap<usize, Tensor> {
+    let mut live: HashMap<usize, Tensor> = schedule
+        .constants
+        .iter()
+        .map(|x| x.1.as_array_constant().map(|z| (*x.0, z.value.clone())))
+        .collect::<Option<HashMap<_, _>>>()
+        .unwrap();
     live.extend(
         schedule
             .scalar_constants
@@ -240,40 +330,16 @@ pub fn evaluate_schedule(schedule: &Schedule) -> HashMap<HashBytes, Tensor> {
     );
     for s in &schedule.instructions {
         match s {
-            Instruction::Compute(circ) => {
-                circ.children().for_each(|x: CircuitRc| {
-                    if !live.contains_key(&x.info().hash) {
-                        println!("FAIL");
-                        dbg!(schedule
-                            .instructions
-                            .iter()
-                            .filter_map(|ins| {
-                                match ins {
-                                    Instruction::Drop(d) => {
-                                        if *d == x.info().hash {
-                                            Some(ins)
-                                        } else {
-                                            None
-                                        }
-                                    }
-                                    Instruction::Compute(c) => {
-                                        if c == &x {
-                                            Some(ins)
-                                        } else {
-                                            None
-                                        }
-                                    }
-                                }
-                            })
-                            .collect::<Vec<&Instruction>>());
-                        circ.compiler_print();
-                        x.compiler_print();
-                        panic!();
+            Instruction::Compute(key, circ) => {
+                let child_keys: Vec<usize> = circ.children().map(get_child_key).collect();
+                child_keys.iter().for_each(|child_key| {
+                    if !live.contains_key(&child_key) {
+                        panic!("FAIL");
                     }
                 });
-                let tensors: Vec<Tensor> = circ
-                    .children()
-                    .map(|x| live[&x.info().hash].clone())
+                let tensors: Vec<Tensor> = child_keys
+                    .iter()
+                    .map(|child_key| live[&child_key].clone())
                     .collect();
                 let result_err = circ.eval_tensors(&tensors, &schedule.device_dtype);
                 if result_err.is_err() {
@@ -282,20 +348,22 @@ pub fn evaluate_schedule(schedule: &Schedule) -> HashMap<HashBytes, Tensor> {
                 }
                 let result = result_err.unwrap();
                 assert!(result.shape()[..] == circ.info().shape[..]);
-                live.insert(circ.info().hash, result);
+                live.insert(*key, result);
             }
-            Instruction::Drop(hash) => {
-                live.remove(hash);
+            Instruction::Drop(key) => {
+                live.remove(key);
             }
         }
     }
     live
 }
 
+/// evaluate a circuit while measuring the numerical scale of tensor contents
+/// and computing "10^10 * x" or such instead of x to avoid numerical overflow
 pub fn evaluate_schedule_adjust_numerical_scale(
     schedule: &Schedule,
     settings: OptimizationSettings,
-) -> HashMap<HashBytes, Tensor> {
+) -> HashMap<usize, Tensor> {
     // we store (tensor, scale) where scale is a number the tensor's been multiplied by
     // so (tensor(1e10),1.0) evaluates to same as (tensor(1),1e-10)
     let mul = |tup: &(Tensor, f64), m: f64| -> (Tensor, f64) {
@@ -332,10 +400,15 @@ pub fn evaluate_schedule_adjust_numerical_scale(
         }
     };
 
-    let mut live: HashMap<HashBytes, (Tensor, f64)> = schedule
+    let mut live: HashMap<usize, (Tensor, f64)> = schedule
         .constants
         .iter()
-        .map(|x| (*x.0, clamp(&(x.1.clone(), 1.0))))
+        .map(|(key, x)| {
+            (
+                *key,
+                clamp(&(x.as_array_constant().unwrap().value.clone(), 1.0)),
+            )
+        })
         .collect();
     live.extend(schedule.scalar_constants.iter().map(|(h, s)| {
         (*h, {
@@ -367,40 +440,15 @@ pub fn evaluate_schedule_adjust_numerical_scale(
     }));
     for s in &schedule.instructions {
         match s {
-            Instruction::Compute(circ) => {
+            Instruction::Compute(key, circ) => {
                 circ.children().for_each(|x: CircuitRc| {
-                    if !live.contains_key(&x.info().hash) {
-                        println!("FAIL");
-                        dbg!(schedule
-                            .instructions
-                            .iter()
-                            .filter_map(|ins| {
-                                match ins {
-                                    Instruction::Drop(d) => {
-                                        if *d == x.info().hash {
-                                            Some(ins)
-                                        } else {
-                                            None
-                                        }
-                                    }
-                                    Instruction::Compute(c) => {
-                                        if c == &x {
-                                            Some(ins)
-                                        } else {
-                                            None
-                                        }
-                                    }
-                                }
-                            })
-                            .collect::<Vec<&Instruction>>());
-                        circ.compiler_print();
-                        x.compiler_print();
-                        panic!();
+                    if !live.contains_key(&get_child_key(x)) {
+                        panic!("FAIL");
                     }
                 });
                 let tensors_and_scales: Vec<(Tensor, f64)> = circ
                     .children()
-                    .map(|x| live[&x.info().hash].clone())
+                    .map(|x| live[&get_child_key(x)].clone())
                     .collect();
                 let tensors: Vec<Tensor> = tensors_and_scales.iter().map(|x| x.0.clone()).collect();
                 let result = match &***circ {
@@ -455,7 +503,7 @@ pub fn evaluate_schedule_adjust_numerical_scale(
                     }
                 };
                 assert!(result.0.shape()[..] == circ.info().shape[..]);
-                live.insert(circ.info().hash, result);
+                live.insert(*key, result);
             }
             Instruction::Drop(hash) => {
                 live.remove(hash);
@@ -471,44 +519,59 @@ pub fn evaluate_schedule_adjust_numerical_scale(
 /// it's assumed you dropped this and are recomputing it
 pub fn order_to_schedule(
     order: &Vec<CircuitRc>,
-    constants: &Vec<ArrayConstant>,
+    constants: &Vec<IrreducibleNode>,
     scalar_constants: &Vec<ScalarConstant>,
     to_keep: HashSet<HashBytes>,
     device_dtype: TorchDeviceDtype,
-) -> Schedule {
+) -> (Schedule, Vec<usize>) {
+    let mut circ_to_id: HashMap<HashBytes, usize> = Default::default();
     let mut result: Schedule = Schedule {
         instructions: vec![],
         constants: constants
             .iter()
-            .map(|x| (x.info().hash, x.value.clone()))
+            .map(|x| {
+                circ_to_id.insert(x.info().hash, circ_to_id.len());
+                (circ_to_id.len() - 1, x.clone())
+            })
             .collect(),
         scalar_constants: scalar_constants
             .iter()
-            .map(|x| (x.info().hash, x.clone()))
+            .map(|x| {
+                circ_to_id.insert(x.info().hash, circ_to_id.len());
+                (circ_to_id.len() - 1, x.clone())
+            })
             .collect(),
         device_dtype,
         split_shapes: None,
         output_circuit: None,
     };
 
-    let mut seen_dependencies: HashSet<HashBytes> = HashSet::new();
+    let mut seen_dependencies: HashSet<usize> = HashSet::new();
     for ex in order.iter().rev() {
         for child in ex.children() {
-            let dep = child.info().hash;
-            if !matches!(
-                &**child,
-                Circuit::ArrayConstant(_) | Circuit::ScalarConstant(_)
-            ) && seen_dependencies.insert(dep)
-                && !to_keep.contains(&dep)
+            let next_id = circ_to_id.len();
+            let dep = *circ_to_id.entry(child.info().hash).or_insert(next_id);
+            if !Leaf::matches(&**child)
+                && seen_dependencies.insert(dep)
+                && !to_keep.contains(&child.info().hash)
             {
                 result.instructions.push(Instruction::Drop(dep));
             }
         }
-        result.instructions.push(Instruction::Compute(ex.clone()));
+        let next_id = circ_to_id.len();
+        let our_id = *circ_to_id.entry(ex.info().hash).or_insert(next_id);
+        let node_here_symbol_children = ex
+            .map_children_unwrap(|child| {
+                child_from_key(circ_to_id[&child.info().hash], child.info().shape.clone()).rc()
+            })
+            .rc();
+        result
+            .instructions
+            .push(Instruction::Compute(our_id, node_here_symbol_children));
         // seen_dependencies.remove(&ex.info().hash);
     }
     result.instructions.reverse();
-    result
+    (result, to_keep.iter().map(|x| circ_to_id[x]).collect())
 }
 
 #[pyclass]
@@ -526,7 +589,8 @@ pub struct Dag {
 }
 impl Dag {
     pub fn get_outputs(&self) -> Vec<usize> {
-        filter_to_idx(&self.parents.iter().collect(), &mut |y| y.is_empty())
+        let result = filter_to_idx(&self.parents.iter().collect(), &mut |y| y.is_empty());
+        result
     }
     /// Create a new dag with a subset of the nodes. Edges to ignored
     /// nodes are replaced with edges that pass through to nodes we're keeping
@@ -620,16 +684,11 @@ pub fn circuits_to_dag(circuits: &[&Circuit]) -> Dag {
     for circuit in circuits {
         visit_circuit(circuit, &mut |c: &Circuit| {
             // ArrayConstants are never added to the dag to begin with because they're always 0 cost
-            if !matches!(c, Circuit::ArrayConstant(_) | Circuit::ScalarConstant(_)) {
+            if !Leaf::matches(c) {
                 let my_number = number_node(c, &mut result);
                 let children_to_consider: Vec<CircuitRc> = c
                     .children()
-                    .filter(|child| {
-                        !matches!(
-                            &***child,
-                            Circuit::ArrayConstant(_) | Circuit::ScalarConstant(_)
-                        )
-                    })
+                    .filter(|child| !Leaf::matches(&***child))
                     .collect();
                 result.children[my_number] = children_to_consider
                     .iter()
@@ -805,37 +864,35 @@ pub fn circuit_to_schedule(
         .iter()
         .map(|x| to_node[&dag.node_hashes[*x]].clone())
         .collect();
-    let mut out = order_to_schedule(
+    let (mut out, kept_keys) = order_to_schedule(
         &circuit_order,
         &to_node
             .iter()
-            .filter_map(|x| x.1.as_array_constant().cloned())
+            .filter_map(|x| Option::<IrreducibleNode>::from((***x.1).clone()))
             .collect(),
         &to_node
             .iter()
             .filter_map(|x| x.1.as_scalar_constant().cloned())
             .collect(),
-        dag.get_outputs()
-            .iter()
-            .map(|x| dag.node_hashes[*x])
-            .collect(),
+        HashSet::from([circuit.info().hash]),
         get_compatible_dtype(&circuit),
     );
-    out.output_circuit = Some(circuit);
-
+    out.output_circuit = Some((kept_keys[0], circuit));
+    let out = replace_module_schedules(&out, context);
     Ok(out)
 }
 
 pub fn circuit_to_schedule_naive_toposort(circuit: CircuitRc) -> Schedule {
+    let circuit = inline_all_modules(circuit);
     let toposorted = toposort_circuit(circuit.clone());
-    let mut result = order_to_schedule(
+    let (mut result, kept_keys) = order_to_schedule(
         &toposorted,
         &vec![],
         &vec![],
         HashSet::from([circuit.info().hash]),
         get_compatible_dtype(&circuit),
     );
-    result.output_circuit = Some(circuit);
+    result.output_circuit = Some((kept_keys[0], circuit));
     result
 }
 
@@ -848,5 +905,132 @@ pub fn scheduled_evaluate(
         circuit.clone(),
         &mut OptimizationContext::new_settings_circuit(settings, circuit.clone()),
     )?;
-    Ok(evaluate_schedule(&schedule)[&circuit.info().hash].clone())
+    Ok(evaluate_schedule(&schedule)[&schedule.output_circuit.unwrap().0].clone())
+}
+
+fn schedule_module_spec(spec: &ModuleNodeSpec, context: &mut OptimizationContext) -> Schedule {
+    context
+        .cache
+        .module_specs_scheduled_same_settings
+        .get(spec)
+        .cloned()
+        .unwrap_or_else(|| {
+            let result = circuit_to_schedule(spec.spec_circuit.clone(), context).unwrap();
+            context
+                .cache
+                .module_specs_scheduled_same_settings
+                .insert(spec.clone(), result.clone());
+            result
+        })
+}
+
+fn replace_module_schedules(schedule: &Schedule, context: &mut OptimizationContext) -> Schedule {
+    let mut result = schedule.clone();
+    let mut scalar_to_id: HashMap<ScalarConstant, usize> = schedule
+        .scalar_constants
+        .iter()
+        .map(|(a, b)| (b.clone(), a.clone()))
+        .collect();
+    let mut irreducible_to_id: HashMap<IrreducibleNode, usize> = schedule
+        .constants
+        .iter()
+        .map(|(a, b)| (b.clone(), a.clone()))
+        .collect();
+    result.instructions = vec![];
+    let mut next_key = schedule.next_key();
+    for ins in &schedule.instructions {
+        match ins {
+            Instruction::Drop(_d) => result.instructions.push(ins.clone()),
+            Instruction::Compute(outermost_k, v) => match &***v {
+                Circuit::ModuleNode(mn) => {
+                    let original_children: Vec<usize> = mn.children().map(get_child_key).collect();
+
+                    let inner_schedule = schedule_module_spec(&mn.spec, context);
+                    if inner_schedule.instructions.is_empty() {
+                        result.instructions.push(Instruction::Compute(
+                            *outermost_k,
+                            identity(
+                                get_child_key(mn.nodes[0].clone()),
+                                mn.nodes[0].info().shape.clone(),
+                            ),
+                        ));
+                    } else {
+                        let mut inner_to_outer_key: HashMap<usize, usize> = HashMap::from([(
+                            inner_schedule.output_circuit.clone().unwrap().0,
+                            *outermost_k,
+                        )]);
+                        for (inner_k, sc) in &inner_schedule.scalar_constants {
+                            let outer = scalar_to_id.get(sc).cloned().unwrap_or_else(|| {
+                                scalar_to_id.insert(sc.clone(), next_key);
+                                result.scalar_constants.insert(next_key, sc.clone());
+                                next_key += 1;
+                                next_key - 1
+                            });
+                            inner_to_outer_key.try_insert(*inner_k, outer).ok();
+                        }
+                        for (inner_k, irreducible) in &inner_schedule.constants {
+                            let outer =
+                                irreducible_to_id
+                                    .get(irreducible)
+                                    .cloned()
+                                    .unwrap_or_else(|| {
+                                        if let Some(pos) =
+                                            mn.spec.input_specs.iter().position(|argspec| {
+                                                argspec.symbol.info().hash
+                                                    == irreducible.info().hash
+                                            })
+                                        {
+                                            original_children[pos]
+                                        } else {
+                                            irreducible_to_id.insert(irreducible.clone(), next_key);
+                                            result.constants.insert(next_key, irreducible.clone());
+                                            next_key += 1;
+                                            next_key - 1
+                                        }
+                                    });
+                            inner_to_outer_key.try_insert(*inner_k, outer).ok();
+                        }
+                        for inner_ins in &inner_schedule.instructions {
+                            match inner_ins {
+                                Instruction::Drop(drop) => {
+                                    result
+                                        .instructions
+                                        .push(Instruction::Drop(inner_to_outer_key[&drop]));
+                                }
+                                Instruction::Compute(inner_k, c) => {
+                                    let new_c = c
+                                        .map_children_unwrap(|child| {
+                                            let child_inner_k = get_child_key(child.clone().rc());
+                                            let child_outer_k =
+                                                inner_to_outer_key.get(&child_inner_k).unwrap();
+                                            child_from_key(
+                                                *child_outer_k,
+                                                child.info().shape.clone(),
+                                            )
+                                            .rc()
+                                        })
+                                        .rc();
+                                    let key = inner_to_outer_key
+                                        .get(&inner_k)
+                                        .cloned()
+                                        .unwrap_or_else(|| {
+                                            inner_to_outer_key.insert(*inner_k, next_key);
+                                            next_key += 1;
+                                            next_key - 1
+                                        });
+                                    result.instructions.push(Instruction::Compute(key, new_c));
+                                    inner_to_outer_key.insert(*inner_k, key);
+                                }
+                            }
+                        }
+                    }
+                }
+                _ => {
+                    result.instructions.push(ins.clone());
+                }
+            },
+        }
+    }
+
+    result
 }
