@@ -2,7 +2,7 @@
 #![feature(map_first_last)]
 // Personally, I like or_fun_call as a lint. But currently code fails...
 #![allow(clippy::too_many_arguments, clippy::or_fun_call)]
-
+#![feature(map_try_insert)]
 use mimalloc::MiMalloc;
 
 #[global_allocator]
@@ -70,7 +70,6 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
         index_split_axes, permute_of_einsum_merge, push_down_index, rearrange_elim_identity,
         rearrange_fuse, rearrange_merge_scalar, remove_add_few_input,
     };
-    use circuit::batching::{batch_inputs_axis_len, batch_largest};
     use circuit::canonicalize::{canonicalize_node_py, deep_canonicalize_py};
     use circuit::circuit_manipulation::{
         deep_map_preorder_py, deep_map_py, filter_nodes_py, path_get, replace_nodes_py,
@@ -81,7 +80,6 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
         cast_circuit, count_nodes, toposort_circuit, total_arrayconstant_size, total_flops,
     };
     use circuit::compiler_heuristics::deep_maybe_distribute_py;
-    use circuit::compiler_strip::strip_names;
     use circuit::concat_rewrite::{
         add_pull_concat, concat_drop_size_zero, concat_fuse, einsum_pull_concat,
         generalfunction_pull_concat, index_concat_drop_unreached, split_to_concat,
@@ -123,6 +121,7 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     m.add_class::<circuit::Concat>()?;
     m.add_class::<circuit::Scatter>()?;
     m.add_class::<circuit::circuit_optimizer::OptimizationSettings>()?;
+    m.add_class::<circuit::circuit_optimizer::OptimizationContext>()?;
     m.add("PyOOMError", py.get_type::<PyOOMError>())?;
 
     m.add_class::<tensor_util::TorchDeviceDtype>()?;
@@ -185,7 +184,11 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     m.add_function(wrap_pyfunction!(deep_canonicalize_py, m)?)?;
     m.add_function(wrap_pyfunction!(canonicalize_node_py, m)?)?;
 
-    m.add_function(wrap_pyfunction!(strip_names, m)?)?;
+    m.add_function(wrap_pyfunction!(circuit::batching::batch_to_concat, m)?)?;
+    m.add_function(wrap_pyfunction!(
+        circuit::compiler_strip::strip_names_py,
+        m
+    )?)?;
     m.add_function(wrap_pyfunction!(deep_maybe_distribute_py, m)?)?;
     m.add_function(wrap_pyfunction!(
         circuit::compiler_heuristics::maybe_distribute_py,
@@ -241,8 +244,6 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     m.add_function(wrap_pyfunction!(deep_push_down_index, m)?)?;
     m.add_function(wrap_pyfunction!(deep_pull_concat_messy, m)?)?;
     m.add_function(wrap_pyfunction!(deep_pull_concat, m)?)?;
-    m.add_function(wrap_pyfunction!(batch_inputs_axis_len, m)?)?;
-    m.add_function(wrap_pyfunction!(batch_largest, m)?)?;
     m.add_function(wrap_pyfunction!(set_named_axes_py, m)?)?;
     m.add_function(wrap_pyfunction!(propagate_named_axes, m)?)?;
     m.add_function(wrap_pyfunction!(toposort_circuit, m)?)?;
