@@ -539,7 +539,8 @@ impl RearrangeSpec {
     #[args(special_case_ones = true)]
     pub fn canonicalize(&self, special_case_ones: bool) -> Self {
         // Remove all indices of size 1
-        let int_sizes: Vec<u64> = self.get_unwrapped_sizes();
+        let int_sizes: Vec<Option<usize>> =
+            self.int_sizes.iter().map(|x| x.clone().into()).collect();
         let mut input_ints = self.input_ints.clone();
         let mut output_ints = self.output_ints.clone();
         if special_case_ones {
@@ -548,7 +549,7 @@ impl RearrangeSpec {
                     .iter()
                     .map(|ints| {
                         ints.iter()
-                            .filter(|i| int_sizes[**i as usize] != 1)
+                            .filter(|i| int_sizes[**i as usize] != Some(1))
                             .copied()
                             .collect()
                     })
@@ -578,11 +579,13 @@ impl RearrangeSpec {
             for (&prev, &curr) in lst.iter().tuple_windows() {
                 let prev_i = prev as usize;
                 let curr_i = curr as usize;
-                if prev_idx[curr_i].is_none() {
-                    prev_idx[curr_i] = Some(Middle(prev));
-                }
-                if next_idx[prev_i].is_none() {
-                    next_idx[prev_i] = Some(Middle(curr));
+                if int_sizes[prev_i].is_some() == int_sizes[curr_i].is_some() {
+                    if prev_idx[curr_i].is_none() {
+                        prev_idx[curr_i] = Some(Middle(prev));
+                    }
+                    if next_idx[prev_i].is_none() {
+                        next_idx[prev_i] = Some(Middle(curr));
+                    }
                 }
                 if prev_idx[curr_i] != Some(Middle(prev)) || next_idx[prev_i] != Some(Middle(curr))
                 {
@@ -611,7 +614,7 @@ impl RearrangeSpec {
         }
 
         let mut map: HashMap<EinInt, EinInt> = HashMap::new();
-        let mut new_int_sizes: Vec<u64> = Vec::new();
+        let mut new_int_sizes: OpShape = sv![];
         let mut already_accounted_shape = vec![false; int_sizes.len()];
 
         let mut merge_tuples_and_renumber_fn = |ints: RInts| {
@@ -625,10 +628,16 @@ impl RearrangeSpec {
                             let new_i: u8 = new_int_sizes.len().try_into().unwrap();
                             map.insert(i, new_i);
                             new_dim.push(new_i);
-                            new_int_sizes.push(int_sizes[i as usize]);
+                            new_int_sizes.push(int_sizes[i as usize].clone().into());
                         } else if !already_accounted_shape[i as usize] {
-                            new_int_sizes[*map.get(&find(i, &mut prev_idx)).unwrap() as usize] *=
-                                int_sizes[i as usize];
+                            if let Some(old_size) = int_sizes[i as usize] {
+                                let new_int_sizes_idx =
+                                    *map.get(&find(i, &mut prev_idx)).unwrap() as usize;
+                                new_int_sizes[new_int_sizes_idx] = Some(
+                                    new_int_sizes[new_int_sizes_idx].unwrap() as usize * old_size,
+                                )
+                                .into();
+                            }
                             already_accounted_shape[i as usize] = true;
                         }
                     }
@@ -643,10 +652,7 @@ impl RearrangeSpec {
         RearrangeSpec {
             input_ints,
             output_ints,
-            int_sizes: new_int_sizes
-                .iter()
-                .map(|&x| OpSize::from(Some(x as usize)))
-                .collect(),
+            int_sizes: new_int_sizes,
         }
     }
 
@@ -773,22 +779,30 @@ impl RearrangeSpec {
                     }
                 })
                 .product();
-            if non_wildcard_size != *l as u64 || !none_indices.is_empty() {
-                if (none_indices.len() == 1 || none_indices.is_empty() && ints.len() == 1 && coerce)
-                    && *l as u64 % non_wildcard_size == 0
-                {
-                    let mut cur_int = ints[0];
-                    if none_indices.len() == 1 {
-                        cur_int = ints[none_indices[0]]
-                    }
-                    let wildcard_size = *l as u64 / non_wildcard_size;
-                    int_sizes[cur_int as usize] = OpSize(wildcard_size);
+            // dbg!(&int_sizes,&non_wildcard_size,&ints,&l);
+            if non_wildcard_size != *l as u64 {
+                if none_indices.len() > 1 {
+                    return Err(TensorInvariantError::RearrangeTooManyWildcardSizes {
+                        spec: self.clone(),
+                    });
+                } else if coerce && ints.len() == 1 {
+                    int_sizes[ints[0] as usize] = OpSize(*l as u64);
+                } else if *l as u64 % non_wildcard_size != 0 {
+                    return Err(TensorInvariantError::RearrangeInputNotConformable {
+                        spec: self.clone(),
+                        shape: shape.clone(),
+                    });
+                } else if none_indices.len() == 1 {
+                    int_sizes[ints[none_indices[0]] as usize] =
+                        OpSize(*l as u64 / non_wildcard_size);
                 } else {
                     return Err(TensorInvariantError::RearrangeInputNotConformable {
                         spec: self.clone(),
                         shape: shape.clone(),
                     });
                 }
+            } else if ints.len() == 1 {
+                int_sizes[ints[0] as usize] = OpSize(*l as u64);
             }
         }
         Ok(Self {
