@@ -43,6 +43,9 @@ pub enum TensorInvariantError {
 
     #[error("RearrangeSpec cannot be converted to no-(), no-squeeze format, {spec:?}=")]
     RearrangeNotConvertable { spec: RearrangeSpec },
+
+    #[error("RearrangeSpec has too many wildcard sizes, {spec:?}")]
+    RearrangeTooManyWildcardSizes { spec: RearrangeSpec },
 }
 
 impl From<TensorInvariantError> for PyErr {
@@ -98,11 +101,11 @@ impl Slice {
         }
     }
 
-    fn start_u(self, l: usize) -> usize {
+    pub fn start_u(self, l: usize) -> usize {
         self.start.map(|x| Self::to_unsigned_loc(x, l)).unwrap_or(0)
     }
 
-    fn stop_u(self, l: usize) -> usize {
+    pub fn stop_u(self, l: usize) -> usize {
         self.stop.map(|x| Self::to_unsigned_loc(x, l)).unwrap_or(l)
     }
 
